import { Frame } from 'react-native-vision-camera';
type DataType = 'uint8' | 'float32';
type OutputArray<T extends DataType> = T extends 'uint8' ? Uint8Array : T extends 'float32' ? Float32Array : never;
interface Size {
    /**
     * The width to resize the Frame to.
     */
    width: number;
    /**
     * The height to resize the Frame to.
     */
    height: number;
}
interface Rect extends Size {
    /**
     * The origin X of the Frame used for cropping. If not set, a center-crop will be performed.
     */
    x: number;
    /**
     * The origin Y of the Frame used for cropping. If not set, a center-crop will be performed.
     */
    y: number;
}
interface Options<T extends DataType> {
    /**
     * Scale the image to the given target size.
     * - If `x` and `y` are set, the plugin will crop to the target origin.
     * - If `x` and `y` are not set, the plugin will perform a center-crop of the image.
     */
    size?: Size | Rect;
    /**
     * Convert the Frame to the given target pixel format.
     *
     * - `'rgb'`: [R, G, B] layout
     * - `'rgba'`: [R, G, B, A]
     * - `'argb'`: [A, R, G, B]
     * - `'bgra'`: [B, G, R, A]
     * - `'bgr'`: [B, G, R]
     * - `'abgr'`: [A, B, G, R]
     */
    pixelFormat: 'rgb' | 'rgba' | 'argb' | 'bgra' | 'bgr' | 'abgr';
    /**
     * The given type to use for the resulting buffer.
     * Each color channel uses this type for representing pixels.
     *
     * - `'uint8'`: Resulting buffer is a `Uint8Array`, values range from 0 to 255
     * - `'float32'`: Resulting buffer is a `Float32Array`, values range from 0.0 to 1.0
     */
    dataType: T;
}
/**
 * An instance of the resize plugin.
 *
 * All temporary memory buffers allocated by the resize plugin
 * will be deleted once this value goes out of scope.
 */
interface ResizePlugin {
    /**
     * Resizes the given Frame to the target width/height and
     * convert it to the given pixel format.
     */
    resize<T extends DataType>(frame: Frame, options: Options<T>): OutputArray<T>;
}
/**
 * Get a new instance of the resize plugin.
 *
 * All temporary memory buffers allocated by the resize plugin
 * will be deleted once the returned value goes out of scope.
 */
export declare function createResizePlugin(): ResizePlugin;
/**
 * Use an instance of the resize plugin.
 *
 * All temporary memory buffers allocated by the resize plugin
 * will be deleted once the component that uses `useResizePlugin()` unmounts.
 */
export declare function useResizePlugin(): ResizePlugin;
export {};
//# sourceMappingURL=index.d.ts.map